/*
    SmartHomeGeeks.io Smart Bathroom Ventilation
    https://smarthomegeeks.io/smart-bathroom-ventilation
*/

#ifdef ESP32
#include <WiFi.h>             // Shelly Plus 2PM
const int switch1 = 5;
const int switch2 = 18;
const int relay1 = 13;
const int relay2 = 12;
#else
#include <ESP8266WiFi.h>      // Shelly 2.5
#include <ESP8266mDNS.h>
const int switch1 = 13;
const int switch2 = 5;
const int relay1 = 4;
const int relay2 = 15;
#endif

#include <WiFiUdp.h>
#include <PubSubClient.h>
#include <ArduinoOTA.h>
#include <ArduinoJson.h>
#include <EEPROM.h>
#include <arduino_secrets.h>

// Device configuration
const String deviceName = "Bathroom Ventilation";

// WiFi configuration
const char* ssid = SECRET_SSID;
const char* password =  SECRET_PASS;

// MQTT configuration
const String mqttClient = "bathroom_ventilation";               // MQTT client name
const String mqttStateTopic = mqttClient + "/state";            // MQTT state topic
const String mqttPctTopic = mqttClient + "/percentage_state";   // MQTT state topic
const String mqttCmdTopic = mqttClient + "/command";            // MQTT command topic
const String mqttAvlTopic = mqttClient + "/availability";       // MQTT availability topic
const char* mqttServer = SECRET_MQTT_BROKER;                    // MQTT broker IP
const char* mqttUser = SECRET_MQTT_USERNAME;                    // MQTT username
const char* mqttPassword = SECRET_MQTT_PASSWORD;                // MQTT password
const int mqttPort = 1883;                                      // MQTT port (default: 1883)

bool prevSwitch1;
bool prevSwitch2;
bool pswitch;
unsigned long lastMQTTconnect = 0;

WiFiClient espClient;
PubSubClient client(espClient);

void setup_wifi() {
  delay(1000);

  // We start by connecting to a WiFi network
  Serial.println();
  Serial.print("Connecting to ");
  Serial.print(ssid);

  WiFi.mode(WIFI_STA);
  WiFi.hostname(mqttClient.c_str());
  WiFi.begin(ssid, password);

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }

  Serial.println();
  Serial.println("WiFi connected");
  Serial.print("IP address: ");
  Serial.println(WiFi.localIP());
}

void autoDiscovery() {
  Serial.println("Publishing Home Assistant Autodiscovery...");

  char buffer[512];

  DynamicJsonDocument deviceData(512);
  deviceData["name"] = deviceName;
  deviceData["mdl"] = "CVE ECO 2P";
  deviceData["ids"] = mqttClient;
  deviceData["mf"] = "SmartHomeGeeks.io for Itho";
  deviceData["sw"] = "2023.08";
  
  // Fan
  DynamicJsonDocument root(512);
  root["name"] = nullptr;
  root["obj_id"] = mqttClient;
  root["stat_t"] = mqttStateTopic;
  root["cmd_t"] = mqttCmdTopic;
  root["spd_rng_max"] = 2;
  root["pct_cmd_t"] = mqttCmdTopic;
  root["pct_stat_t"] = mqttPctTopic;
  root["uniq_id"] = "fan." + mqttClient;
  root["avty_t"] = mqttAvlTopic;
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/fan/" + mqttClient + "/config").c_str(), buffer, true);
  root.clear();

  // Switch
  root["name"] = "Physical Switch";
  root["obj_id"] = mqttClient + "_pswitch";
  root["ic"] = "mdi:knob";
  root["uniq_id"] = "switch." + mqttClient;
  root["stat_t"] = mqttClient + "/pswitchStatus";
  root["cmd_t"] = mqttClient + "/pswitchCmd";
  root["avty_t"] = mqttClient + "/availability";
  root["ent_cat"] = "config";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/switch/" + mqttClient + "/config").c_str(), buffer, true);
  root.clear();

  // RSSI sensor
  root["name"] = "RSSI";
  root["obj_id"] = mqttClient + "_rssi";
  root["stat_t"] = mqttClient + "/rssi";
  root["unit_of_meas"] = "dBm";
  root["ic"] = "mdi:access-point";
  root["uniq_id"] = "sensor." + mqttClient + "_rssi";
  root["avty_t"] = mqttAvlTopic;
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/rssi/config").c_str(), buffer, true);
  root.clear();

  // IP
  root["name"] = "IP";
  root["obj_id"] = mqttClient + "_ip";
  root["stat_t"] = mqttClient + "/ip";
  root["ic"] = "mdi:ip-network";
  root["uniq_id"] = "sensor." + mqttClient + "_ip";
  root["avty_t"] = mqttClient + "/availability";
  root["ent_cat"] = "diagnostic";
  for (JsonPair kv : deviceData.as<JsonObject>()) root["dev"][kv.key()] = kv.value();
  serializeJson(root, buffer);
  client.publish(("homeassistant/sensor/" + mqttClient + "/ip/config").c_str(), buffer, true);
  root.clear();
}

void connectmqtt() {
  // prevent constant reconnect
  if (millis() - lastMQTTconnect < 5000) return;
  lastMQTTconnect = millis();

  if (client.connect(mqttClient.c_str(), mqttUser, mqttPassword, mqttAvlTopic.c_str(), 1, true, "offline")) {
    Serial.println("connected to MQTT");
    client.setBufferSize(512);
    autoDiscovery();

    client.subscribe((mqttClient + "/command").c_str());
    client.subscribe((mqttClient + "/pswitchCmd").c_str());
    client.publish(mqttAvlTopic.c_str(), "online", true);
    client.publish((mqttClient + "/ip").c_str(),  WiFi.localIP().toString().c_str(), true);
  } else {
    Serial.print("failed, rc=");
    Serial.print(client.state());
    Serial.println(" try again in 5 seconds");
  }
}

void setup() {
  Serial.begin(115200);
  pinMode(relay1, OUTPUT);
  pinMode(relay2, OUTPUT);

  // Setup WiFi  
  setup_wifi();

  // Configure OTA
  ArduinoOTA.setHostname(mqttClient.c_str());
  ArduinoOTA.setPassword(SECRET_OTA_PASSWORD);
  ArduinoOTA.begin(); 

  //Connect MQTT
  client.setServer(mqttServer, mqttPort);
  client.setCallback(callback);

  EEPROM.begin(1);
  pswitch = EEPROM.read(0);
  client.publish((mqttClient + "/chimeStatus").c_str(), (pswitch ? "ON" : "OFF"), true);
}

void callback(char* topic, byte* payload, unsigned int length) {
  Serial.print("Message arrived: ");
  for (int i=0;i<length;i++) {
    Serial.print((char)payload[i]);
  }
  Serial.println();

  if (String(topic) == (mqttClient + "/pswitchCmd")) {
    pswitch = (!strncmp((char*)payload, "ON", length) ? true : false);
    client.publish((mqttClient + "/pswitchStatus").c_str(), (!strncmp((char*)payload, "ON", length) ? "ON" : "OFF"), true);
    EEPROM.write(0, pswitch);
    EEPROM.commit();
  } else if (String(topic) == (mqttClient + "/command"))  {
    if (!strncmp((char*)payload, "1", length)) setFanSpeed("1");
    else if (!strncmp((char*)payload, "2", length) || !strncmp((char*)payload, "ON", length)) setFanSpeed("2");
    else setFanSpeed("0");
  }
}

void setFanSpeed(String speed) {
  Serial.print("Changing fan speed to ");
  Serial.println(speed);
  digitalWrite(relay1, LOW);
  digitalWrite(relay2, LOW);
  if (speed == "1") digitalWrite(relay1, HIGH);
  else if (speed == "2") digitalWrite(relay2, HIGH);

  client.publish(mqttStateTopic.c_str(), (speed == "0" ? "OFF" : "ON"), true);
  client.publish(mqttPctTopic.c_str(), speed.c_str(), true);
}

void loop() {
  if (!client.connected()) connectmqtt();

  if (prevSwitch1 != digitalRead(switch1) || prevSwitch2 != digitalRead(switch2)) {
    Serial.println("Switch change detected");
    if (pswitch) {
      if (digitalRead(switch1)) setFanSpeed("1");
      else if (digitalRead(switch2)) setFanSpeed("2");
      else setFanSpeed("0");
    }

    // update RSSI
    client.publish((mqttClient + "/rssi").c_str(),  String(WiFi.RSSI()).c_str(), true);
  }

  prevSwitch1 = digitalRead(switch1);
  prevSwitch2 = digitalRead(switch2);

  delay(100);

  ArduinoOTA.handle();

  client.loop();
}
